import warnings
from pkg_resources import resource_filename

from tqdm import tqdm
import numpy as np
import pandas as pd
from cyvcf2 import Writer, VCF
from keras import backend as K
from keras.models import load_model
from sklearn.externals import joblib
from concise.preprocessing import encodeDNA

from .generic import logit
from .utils.postproc import transform


ACCEPTOR_INTRON = resource_filename('mmsplice', 'models/Intron3.h5')
DONOR = resource_filename('mmsplice', 'models/Donor.h5')
EXON = resource_filename('mmsplice', 'models/Exon.h5')
EXON3 = resource_filename('mmsplice', 'models/Exon_prime3.h5')
ACCEPTOR = resource_filename('mmsplice', 'models/Acceptor.h5')
DONOR_INTRON = resource_filename('mmsplice', 'models/Intron5.h5')
LINEAR_MODEL = joblib.load(resource_filename(
    'mmsplice', 'models/linear_model.pkl'))
LOGISTIC_MODEL = joblib.load(resource_filename(
    'mmsplice', 'models/Pathogenicity.pkl'))
EFFICIENCY_MODEL = joblib.load(resource_filename(
    'mmsplice', 'models/splicing_efficiency.pkl'))


class MMSplice(object):
    """ Load modules of mmsplice model, perform prediction with inputs come from a dataloader.

    Args:
        acceptor_intronM: acceptor intron model, score acceptor intron sequence.
        acceptorM: accetpor splice site model. Score acceptor sequence with 50bp from intron, 3bp from exon.
        exonM: exon model, score exon sequence.
        donorM: donor splice site model, score donor sequence with 13bp in the intron, 5bp in the exon.
        donor_intronM: donor intron model, score donor intron sequence.
        exon_cut_l: when extract exon feature, how many base pair to cut out at the begining of an exon
        exon_cut_r: when extract exon feature, how many base pair to cut out at the end of an exon
           (cut out the part that is considered as acceptor site or donor site)
        acceptor_intron_cut: how many bp to cut out at the end of acceptor intron that consider as acceptor site
        donor_intron_cut: how many bp to cut out at the end of donor intron that consider as donor site
        acceptor_intron_len: what length in acceptor intron to consider for acceptor site model
        acceptor_exon_len: what length in acceptor exon to consider for acceptor site model
        donor_intron_len: what length in donor intron to consider for donor site model
        donor_exon_len: what length in donor exon to consider for donor site model
    """

    def __init__(self,
                 acceptor_intronM=ACCEPTOR_INTRON,
                 acceptorM=ACCEPTOR,
                 exonM=EXON,
                 donorM=DONOR,
                 donor_intronM=DONOR_INTRON,

                 # parameters to split the sequence
                 exon_cut_l=0,
                 exon_cut_r=0,
                 acceptor_intron_cut=6,
                 donor_intron_cut=6,
                 acceptor_intron_len=50,
                 acceptor_exon_len=3,
                 donor_exon_len=5,
                 donor_intron_len=13
                 ):

        K.clear_session()
        self.exon_cut_l = exon_cut_l
        self.exon_cut_r = exon_cut_r
        self.acceptor_intron_cut = acceptor_intron_cut
        self.donor_intron_cut = donor_intron_cut
        self.acceptor_intron_len = acceptor_intron_len
        self.acceptor_exon_len = acceptor_exon_len
        self.donor_exon_len = donor_exon_len
        self.donor_intron_len = donor_intron_len

        self.acceptor_intronM = load_model(acceptor_intronM)
        self.acceptorM = load_model(acceptorM)
        self.exonM = load_model(exonM)
        self.donorM = load_model(donorM)
        self.donor_intronM = load_model(donor_intronM)

    def predict_on_batch(self, x, **kwargs):
        ''' Use when load batch with sequence already splited.
        This way various length sequences are padded to the same length.
        x is a batch with sequence not encoded.
        Need to be encoded here for collate function to work
        '''
        fts = x['seq']
        acceptor_intron = encodeDNA(
            fts['acceptor_intron'].tolist(), seq_align="end")
        acceptor = encodeDNA(fts['acceptor'].tolist(), seq_align="end")
        exon = encodeDNA(fts['exon'].tolist(), seq_align="end")
        donor = encodeDNA(fts['donor'].tolist(), seq_align="end")
        donor_intron = encodeDNA(fts['donor_intron'].tolist(), seq_align="end")
        score = np.concatenate([
            self.acceptor_intronM.predict(acceptor_intron),
            logit(self.acceptorM.predict(acceptor)),
            self.exonM.predict(exon),
            logit(self.donorM.predict(donor)),
            self.donor_intronM.predict(donor_intron)], axis=1)

        return score

    def predict_on_unsplitted_batch(self, inputs):
        ''' Apply self.predict function to a batch. Apply to cases where sequences
        are not splited in the dataloader. Variable input lengths are not padded to the same,
        but kept the same.
        Each loop split sequence and apply the model.
        '''
        dt = pd.DataFrame.from_dict(inputs)
        return dt.apply(self.predict, axis=1).values

    def predict(self, x):
        ''' Use when incomming sequence is not splited.
        In this way, sequence with different length are one-hot encoded to the same length through padding.
        x: a sequence to split
        '''
        seq = x['seq']
        intronl_len = x['intronl_len']
        intronr_len = x['intronr_len']

        if isinstance(seq, dict):
            fts = seq
        else:
            fts = self.split(seq, intronl_len, intronr_len)

        score = pd.Series([
            self.acceptor_intronM.predict(fts['acceptor_intron'])[0][0],
            logit(self.acceptorM.predict(fts['acceptor']))[0][0],
            self.exonM.predict(fts['exon'])[0][0],
            logit(self.donorM.predict(fts['donor']))[0][0],
            self.donor_intronM.predict(fts['donor_intron'])[0][0]])

        return score

    def split(self, x, intronl_len=100, intronr_len=80):
        ''' x: a sequence to split
        '''
        lackl = self.acceptor_intron_len - \
            intronl_len  # need to pad N if left seq not enough long
        if lackl >= 0:
            x = "N" * (lackl + 1) + x
            intronl_len += lackl + 1
        lackr = self.donor_intron_len - intronr_len
        if lackr >= 0:
            x = x + "N" * (lackr + 1)
            intronr_len += lackr + 1

        acceptor_intron = x[:intronl_len - self.acceptor_intron_cut]
        acceptor = x[(intronl_len - self.acceptor_intron_len):(intronl_len + self.acceptor_exon_len)]
        exon = x[(intronl_len + self.exon_cut_l):(-intronr_len - self.exon_cut_r)]
        donor = x[(-intronr_len - self.donor_exon_len):(-intronr_len + self.donor_intron_len)]
        donor_intron = x[-intronr_len + self.donor_intron_cut:]

        if donor[self.donor_exon_len:self.donor_exon_len + 2] != "GT":
            warnings.warn("None GT donor", UserWarning)
        if acceptor[self.acceptor_intron_len - 2:self.acceptor_intron_len] != "AG":
            warnings.warn("None AG donor", UserWarning)
        if len(exon) == 0:
            exon = 'N'

        return {
            "acceptor_intron": encodeDNA([acceptor_intron]),
            "acceptor": encodeDNA([acceptor]),
            "exon": encodeDNA([exon]),
            "donor": encodeDNA([donor]),
            "donor_intron": encodeDNA([donor_intron])
        }


def predict_all(model, dataloader, batch_size=256, assembly_fn='sum'):
    ''' TODO: multithreading
    TODO: pack into an object that has method convert predictions to a dataframe
    '''
    ID = []
    predicts = []
    exons = []
    dt_iter = dataloader.batch_iter(batch_size=batch_size)
    for x in dt_iter:
        ref_pred = model.predict_on_batch(x['inputs'])
        alt_pred = model.predict_on_batch(x['inputs_mut'])
        ID.append(x['metadata']['variant']['ID'])
        exons.append(x['metadata']['annotation'])
        pred = assembly_fn(alt_pred - ref_pred)
        predicts.append(pred.astype(str))
    ID = np.concatenate(ID)
    predicts = np.concatenate(predicts)
    exons = np.concatenate(exons)
    predicts = np.core.defchararray.add(exons, predicts)
    # predictions = dict(zip(ID, predicts))
    predictions = pd.DataFrame({'ID': ID, 'predicts': predicts})
    predictions = predictions.groupby(
        ['ID'], as_index=False).agg(lambda x: ','.join(set(x)))
    predictions = dict(zip(predictions.ID, predictions.predicts))

    return predictions


def predict_all_table(model,
                      dataloader,
                      batch_size=512,
                      assembly=True,
                      split_seq=True,
                      progress=True,
                      # assembly_fn=LINEAR_MODEL,
                      pathogenicity=False,
                      splicing_efficiency=False):
    ''' Return the prediction as a table
        exon_scale_factor: can be determined through cross validation.
        Args:
            The mmsplice model object
            dataloader: dataloader object.
            split_seq: is the input sequence from dataloader splited?
            progress: show progress bar?
            assembly_fn: function to assemble modular predictions.
            pathogenicity: whether output prediction of pathogenicity
            splicing_efficiency: whether output prediction of splicing_efficiency
    '''
    ID = []
    ref_pred = []
    alt_pred = []
    exons = []
    dt_iter = dataloader.batch_iter(batch_size=batch_size)
    if progress:
        dt_iter = tqdm(dt_iter)
    for x in dt_iter:
        if split_seq:
            ref_pred.append(model.predict_on_batch(x['inputs']))
            alt_pred.append(model.predict_on_batch(x['inputs_mut']))
        else:
            ref_pred.append(model.predict_on_unsplitted_batch(x['inputs']))
            alt_pred.append(model.predict_on_unsplitted_batch(x['inputs_mut']))
        ID.append(x['metadata']['variant']['STR'])
        exons.append(x['metadata']['annotation'])
    ID = np.concatenate(ID)
    ref_pred = np.concatenate(ref_pred)
    alt_pred = np.concatenate(alt_pred)
    ref_pred = pd.DataFrame(ref_pred, columns=['mmsplice_ref_acceptorIntron',
                                               'mmsplice_ref_acceptor',
                                               'mmsplice_ref_exon',
                                               'mmsplice_ref_donor',
                                               'mmsplice_ref_donorIntron'])
    alt_pred = pd.DataFrame(alt_pred, columns=['mmsplice_alt_acceptorIntron',
                                               'mmsplice_alt_acceptor',
                                               'mmsplice_alt_exon',
                                               'mmsplice_alt_donor',
                                               'mmsplice_alt_donorIntron'])
    exons = np.concatenate(exons)
    pred = pd.DataFrame({'ID': ID, 'exons': exons})
    if assembly:
        ref_pred = ref_pred.values
        alt_pred = alt_pred.values
        X = alt_pred - ref_pred
        if pathogenicity:
            X_pathogenicity = transform(X, region_only=True)
            # design matrix for logistic model to predict pathogenicity
            X_pathogenicity = np.concatenate([ref_pred, alt_pred, X_pathogenicity[:, -3:]], axis=-1)
            delt_pred = LOGISTIC_MODEL.predict_proba(X_pathogenicity)[:, 1]
            pred['mmsplice_pathogenicity'] = delt_pred
        if splicing_efficiency:
            X_splicing_efficiency = transform(X, region_only=False)
            X_splicing_efficiency = X_splicing_efficiency[:,[1,2,3,5]] # no intronic modules
            delt_pred = EFFICIENCY_MODEL.predict(X_splicing_efficiency)
            pred['mmsplice_dse'] = delt_pred
        X = transform(X, region_only=False)
        delt_pred = LINEAR_MODEL.predict(X)
        pred['mmsplice_dlogitPsi'] = delt_pred
    else:
        pred = pd.concat([pred, ref_pred, alt_pred], axis=1)
    return pred


def writeVCF(vcf_in, vcf_out, predictions):
    with VCF(vcf_in) as vcf:
        vcf.add_info_to_header({
            'ID': 'mmsplice',
            'Description': 'mmsplice splice variant effect',
            'Type': 'Character',
            'Number': '.'
        })

        with Writer(vcf_out, vcf) as w:
            for var in vcf:
                pred = predictions.get(var.ID)
                if pred is not None:
                    var.INFO['mmsplice'] = pred
                w.write_record(var)



from .postproc import max_varEff, \
    transform, \
    predict_deltaLogitPsi, \
    predict_pathogenicity

__all__ = [
    'max_varEff',
    'transform',
    'predict_deltaLogitPsi',
    'predict_pathogenicity'
]



from gevent import monkey
monkey.patch_all()

import click
from gevent.pywsgi import WSGIServer
from mmsplice.api import app


@click.group()
def cli():
    pass


@cli.command(name='run-api')
@click.option('--port', default=5000, help='Port of python server')
def run_api(port):
    http_server = WSGIServer(('', port), app.wsgi_app)
    http_server.serve_forever()


if __name__ == '__main__':
    cli()



import pickle
import warnings
import functools

import gffutils
from cyvcf2 import VCF
from pyfaidx import Fasta
from kipoi.data import SampleIterator
from kipoi.metadata import GenomicRanges
from concise.preprocessing import encodeDNA

from .generic import Variant, get_var_side
from .interval_tree import IntervalTree, Interval


class VariantInterval(Interval):

    def __init__(self, REF=None, ALT=None, **kwargs):
        super().__init__(**kwargs)
        self.REF = REF
        self.ALT = ALT

    @classmethod
    def from_Variant(cls, variant):
        # Only support one alternative.
        # If multiple alternative, need to split into multiple variants
        if len(variant.ALT) != 1:
            warnings.warn('%s has more than one alternative sequence,'
                         'only the first one was taken,'
                         'split into mutliple variants with bedtools' % variant.__repr__(), UserWarning)
        if variant.ID is None:
            ID = '.'
        else:
            ID = variant.ID
        return cls(chrom=variant.CHROM,
                   start=variant.POS,
                   end=variant.POS,
                   strand="*",
                   name=ID,
                   REF=variant.REF,
                   ALT=variant.ALT[0])

    def to_Variant(self, strand, side):
        # convert to my Variant class, which needs strand and side
        assert self.start == self.end
        var = Variant(CHROM=self.chrom,
                      POS=self.start,
                      REF=self.REF,
                      ALT=self.ALT,
                      strand=strand,
                      ID=self.name)
        var.side = side
        return var


class ExonInterval(gffutils.Feature):

    def __init__(self, order=-1, **kwargs):
        super().__init__(**kwargs)
        self.order = order
        self.name = self.attributes["exon_id"][0]
        self.transcript_id = self.attributes["transcript_id"][0]
        self.gene_id = self.attributes["gene_id"][0]
        # self.biotype = self.attributes["gene_biotype"][0]
        self.Exon_Start = self.start  # with overhang, will be set later
        self.Exon_End = self.end  # with overhang
        self.overhang = (0, 0)
        self._isLast = False

    @property
    def isLast(self):
        return self._isLast

    @isLast.setter
    def isLast(self, value):
        self._isLast = value

    @property
    def isFirst(self):
        return self.order == 1

    @property
    def grange(self):
        return GenomicRanges(self.chrom,
                             self.start,
                             self.end,
                             self.transcript_id,
                             self.strand)

    def __str__(self):
        return '{0}_{1}_{2}:{3}'.format(
            self.chrom, self.Exon_Start, self.Exon_End, self.strand)

    @property
    def to_dict(self):
        return {'isLast': self.isLast,
                'isFirst': self.isFirst,
                'order': self.order,
                'name': self.name,
                'gene_id': self.gene_id,
                'Exon_Start': self.Exon_Start,
                'Exon_End': self.Exon_End,
                'intronl_len': self.overhang[0],
                'intronr_len': self.overhang[1],
                'seqid': self.seqid,
                'strand': self.strand,
                'start': self.start,
                'end': self.end}

    @classmethod
    def from_Feature(cls,
                     feature,
                     overhang=(0, 0)):
        # convert gffutils.Feature to ExonInterval
        iv = cls(seqid=feature.chrom,
                 source=feature.source,
                 start=feature.start,  # exon start
                 end=feature.end,  # exon end
                 strand=feature.strand,
                 frame=feature.frame,
                 attributes=feature.attributes,
                 order=int(feature.attributes['exon_number'][0]))
        if iv.strand == "+":
            iv.start = iv.Exon_Start - overhang[0]
            iv.end = iv.Exon_End + overhang[1]
        else:
            iv.start = iv.Exon_Start - overhang[1]
            iv.end = iv.Exon_End + overhang[0]
        iv.overhang = overhang
        return iv

    @classmethod
    def from_exonfile(cls, exon, attributes, overhang=(0, 0)):
        iv = cls(seqid=exon.CHROM,
                 source='',
                 start=exon.Exon_Start,  # exon start
                 end=exon.Exon_End,  # exon end
                 strand=exon.strand,
                 frame='',
                 attributes=attributes,
                 order=attributes['order'])
        if iv.strand == "+":
            iv.start = iv.Exon_Start - overhang[0]
            iv.end = iv.Exon_End + overhang[1]
        else:
            iv.start = iv.Exon_Start - overhang[1]
            iv.end = iv.Exon_End + overhang[0]
        iv.overhang = overhang
        return iv

    def get_seq(self, fasta, use_strand=True):
        seq = self.sequence(fasta, use_strand=use_strand)
        seq = seq.upper()
        return seq

    def _ref_check(self, fasta, variant):
        '''Checks that ref seq matchs with fasta'''
        ref_check = fasta.get_seq(
            self.chrom,
            variant.POS, variant.POS +
            len(variant.REF) - 1,  # -1 because 1 based
            self.strand == '-').seq  # last option corresponds to rc= in Fasta.get_seq function from pyfaidx
        return ref_check.upper() != variant.REF

    def _get_snp_seq(self, seq, variant, position):
        '''Seq of SNP, or more than one nt equal length substitution'''
        mut_seq = seq[:max(0, position)] + variant.ALT + \
            seq[position + len(variant.ALT):]
        if len(mut_seq) > len(seq):
            mut_seq = mut_seq[:len(seq)]
            assert len(mut_seq) == len(seq), variant
        return mut_seq

    def _get_insertion_seq(self, seq, variant, position):
        # insertion
        if len(variant.REF) + position > len(seq):
            # The actual variant position exceeded the retrieved sequence
            return seq
        mut_seq = seq[:max(0, position)] + variant.ALT + \
            seq[position + len(variant.REF):]
        assert len(seq) - len(mut_seq) == variant.len_diff, variant
        if variant.side is None:
            # only substitute
            return mut_seq
        elif variant.side == 'left':
            return mut_seq[-variant.len_diff:]
        else:
            return mut_seq[:variant.len_diff]

    def _get_del_seq(self, seq, variant, position, fasta):
            # deletion
        if len(variant.REF) + position > len(seq):
            # The actual variant position exceeded the retrieved sequence
            return seq
        if variant.side is None:
            mut_seq = seq[:max(0, position)] + variant.ALT + \
                seq[position + len(variant.REF):]
            assert len(seq) - len(mut_seq) == variant.len_diff, variant
            return mut_seq
        elif variant.side == 'left':
            if self.strand == "+":
                mut_seq = fasta.get_seq(
                    self.chrom,
                    self.start - variant.len_diff, self.end,
                    self.strand == '-').seq
            else:
                mut_seq = fasta.get_seq(
                    self.chrom,
                    self.start, self.end + variant.len_diff,
                    self.strand == '-').seq
            mut_seq = mut_seq[:position + variant.len_diff] + \
                variant.ALT + seq[position + len(variant.REF):]
            assert len(mut_seq) == len(seq), variant
            return mut_seq
        else:
            if self.strand == "+":
                mut_seq = fasta.get_seq(
                    self.chrom,
                    self.start, self.end + variant.len_diff,
                    self.strand == '-').seq
            else:
                mut_seq = fasta.get_seq(
                    self.chrom,
                    self.start - variant.len_diff, self.end,
                    self.strand == '-').seq
            mut_seq = seq[:position] + variant.ALT + \
                mut_seq[position + len(variant.REF):]
            assert len(mut_seq) == len(seq), variant
            return mut_seq

    def get_mut_seq(self, fasta, variant):
        assert variant.side in (None, "left", "right")

        if self._ref_check(fasta, variant):
            warnings.warn(
                "Reference not match, cannot mutate, return original sequence.", UserWarning)
            print(variant)
            return self.get_seq(fasta)

        seq = self.get_seq(fasta)
        position = self._var_pos(variant)

        # position is 1 based, len(seq) 0 based
        if position < 0 or position >= len(seq):
            return seq
        elif variant.len_diff == 0:
            return self._get_snp_seq(seq, variant, position)
        elif variant.len_diff < 0:
            return self._get_insertion_seq(seq, variant, position)
        elif variant.len_diff > 0:
            return self._get_del_seq(seq, variant, position, fasta)

    def _var_pos(self, variant):
        """ Get variant relative position in the sequence
        """
        if self.strand == "+":
            return variant.POS - self.start
        else:
            return self.end - variant.POS - len(variant.REF) + 1


class FastaSeq(Fasta):
    ''' Implement a getSeq method that return upper string and take strand
    '''

    def getSeq(self, iv):
        seq = self.get_seq(iv.chrom, iv.start, iv.end, iv.strand == '-')
        return seq.seq.upper()


@functools.lru_cache(maxsize=1)
def GenerateExonIntervalTree(gtf_file,
                             overhang=(100, 100),  # overhang from the exon
                             gtf_db_path=":memory:",
                             out_file=None,
                             disable_infer_transcripts=True,
                             disable_infer_genes=True,
                             firstLastNoExtend=True,
                             source_filter=None):
    """
    Build IntervalTree object from gtf file for one feature unit (e.g. gene, exon). If give out_file, pickle it.
    Args:
        gtf_file: gtf format file or pickled Intervaltree object.
        overhang: flanking intron length to take along with exon. Corresponding to left (acceptor side) and right (donor side)
        gtf_db_path: (optional) gtf database path. Database for one gtf file only need to be created once
        out_file: (optional) file path to store the pickled Intervaltree obejct. Next time run it can be given to `gtf_file`
        disable_infer_transcripts: option to disable infering transcripts. Can be True if the gtf file has transcripts annotated.
        disable_infer_genes: option to disable infering genes. Can be True if the gtf file has genes annotated.
        firstLastNoExtend: if True, overhang is not taken for 5' of the first exon, or 3' of the last exon of a gene.
        source_filter: gene source filters, such as "protein_coding" filter for protein coding genes
    """
    try:
        gtf_db = gffutils.interface.FeatureDB(gtf_db_path)
    except ValueError:
        gtf_db = gffutils.create_db(
            gtf_file,
            gtf_db_path,
            disable_infer_transcripts=disable_infer_transcripts,
            disable_infer_genes=disable_infer_genes)

    genes = gtf_db.features_of_type('gene')
    exonTree = IntervalTree()
    default_overhang = overhang
    for gene in genes:
        if source_filter is not None:
            if gene.source != source_filter:
                continue
        for exon in gtf_db.children(gene, featuretype='exon'):
            isLast = False  # track whether is last exon
            if firstLastNoExtend:
                if (gene.strand == "+" and exon.end == gene.end) or (gene.strand == "-" and exon.start == gene.start):
                    overhang = (overhang[0], 0)
                    isLast = True
                elif (gene.strand == "+" and exon.start == gene.start) or (gene.strand == "-" and exon.end == gene.end):
                    # int(exon.attributes['exon_number'][0]) == 1:
                    overhang = (0, overhang[1])
            iv = ExonInterval.from_Feature(exon, overhang)
            iv.isLast = isLast
            overhang = default_overhang
            exonTree.insert(iv)
    if out_file is not None:
        with open(out_file, 'wb') as f:
            pickle.dump(exonTree, f)
    return exonTree


class SplicingVCFDataloader(SampleIterator):
    """
    Load genome annotation (gtf) file along with a vcf file, return wt sequence and mut sequence.
    Args:
        gtf: gtf file or pickled gtf IntervalTree. Can be dowloaded from ensembl/gencode. Filter for protein coding genes.
        fasta_file: file path; Genome sequence
        vcf_file: vcf file, each line should contain one and only one variant, left-normalized
        spit_seq: whether or not already split the sequence when loading the data. Otherwise it can be done in the model class.
        endcode: if split sequence, should it be one-hot-encoded
        exon_cut_l: when extract exon feature, how many base pair to cut out at the begining of an exon
        exon_cut_r: when extract exon feature, how many base pair to cut out at the end of an exon
           (cut out the part that is considered as acceptor site or donor site)
        acceptor_intron_cut: how many bp to cut out at the end of acceptor intron that consider as acceptor site
        donor_intron_cut: how many bp to cut out at the end of donor intron that consider as donor site
        acceptor_intron_len: what length in acceptor intron to consider for acceptor site model
        acceptor_exon_len: what length in acceptor exon to consider for acceptor site model
        donor_intron_len: what length in donor intron to consider for donor site model
        donor_exon_len: what length in donor exon to consider for donor site model
        out_file: file path to save pickle file for IntervalTree object derived from GTF annotation. 
            Once the IntervalTree object is generated and saved as a pickle file, next run it can be directly provided to the `gtf` argument,
            the program will save time by not generating it again.
        variant_filter: if set True (default), variants with `FILTER` field other than `PASS` will be filtered out.
        **kwargs: kwargs for `GenerateExonIntervalTree` object
    """

    def __init__(self,
                 gtf,
                 fasta_file,
                 vcf_file=None,
                 split_seq=False,
                 exon_cut_l=0,
                 exon_cut_r=0,
                 acceptor_intron_cut=6,
                 donor_intron_cut=6,
                 acceptor_intron_len=50,
                 acceptor_exon_len=3,
                 donor_exon_len=5,
                 donor_intron_len=13,
                 variant_filter=True,
                 encode=True,
                 **kwargs):
        try:
            with open(gtf, 'rb') as f:
                self.exons = pickle.load(f)
        except (FileNotFoundError, pickle.UnpicklingError):
            self.exons = GenerateExonIntervalTree(gtf, **kwargs)
        import six
        if isinstance(fasta_file, six.string_types):
            fasta = Fasta(fasta_file, as_raw=False)
        self.fasta = fasta
        self.ssGenerator = self.spliceSiteGenerator(vcf_file, self.exons, variant_filter)

        self.encode = encode
        self.split_seq = split_seq
        self.exon_cut_l = exon_cut_l
        self.exon_cut_r = exon_cut_r
        self.acceptor_intron_cut = acceptor_intron_cut
        self.donor_intron_cut = donor_intron_cut
        self.acceptor_intron_len = acceptor_intron_len
        self.acceptor_exon_len = acceptor_exon_len
        self.donor_exon_len = donor_exon_len
        self.donor_intron_len = donor_intron_len

    @staticmethod
    def spliceSiteGenerator(vcf_file, exonTree, variant_filter=True):
        variants = VCF(vcf_file)
        for var in variants:
            if variant_filter and var.FILTER:
                next
            iv = VariantInterval.from_Variant(var)

            matches = map(lambda x: x.interval,
                          exonTree.intersect(iv, ignore_strand=True))

            for match in matches:
                side = get_var_side((
                    var.POS,
                    var.REF,
                    var.ALT,
                    match.Exon_Start,
                    match.Exon_End,
                    match.strand
                ))
                var = iv.to_Variant(match.strand, side)  # to my Variant class
                yield match, var

    def __iter__(self):
        return self

    def __next__(self):
        ss, var = next(self.ssGenerator)
        seq = ss.get_seq(self.fasta).upper()
        mut_seq = ss.get_mut_seq(self.fasta, var).upper()

        if self.split_seq:
            seq = self.split(seq, ss.overhang)
            mut_seq = self.split(mut_seq, ss.overhang)

        return {
            'inputs': {
                'seq': seq,
                'intronl_len': ss.overhang[0],
                'intronr_len': ss.overhang[1],

            },
            'inputs_mut': {
                'seq': mut_seq,
                'intronl_len': ss.overhang[0],
                'intronr_len': ss.overhang[1]
            },
            'metadata': {
                'ranges': ss.grange,
                'variant': var.to_dict,
                'ExonInterval': ss.to_dict,
                'annotation': str(ss)
            }
        }

    def batch_predict_iter(self, **kwargs):
        """Returns samples directly useful for prediction x["inputs"]
        Args:
          **kwargs: Arguments passed to self.batch_iter(**kwargs)
        """
        return (x for x in self.batch_iter(**kwargs))

    def split(self, x, overhang):
        ''' x: a sequence to split
        '''
        intronl_len, intronr_len = overhang
        # need to pad N if left seq not enough long
        lackl = self.acceptor_intron_len - intronl_len
        if lackl >= 0:
            x = "N" * (lackl + 1) + x
            intronl_len += lackl + 1
        lackr = self.donor_intron_len - intronr_len
        if lackr >= 0:
            x = x + "N" * (lackr + 1)
            intronr_len += lackr + 1

        acceptor_intron = x[:intronl_len - self.acceptor_intron_cut]

        acceptor_start = intronl_len - self.acceptor_intron_len
        acceptor_end = intronl_len + self.acceptor_exon_len
        acceptor = x[acceptor_start: acceptor_end]

        exon_start = intronl_len + self.exon_cut_l
        exon_end = -intronr_len - self.exon_cut_r
        exon = x[exon_start: exon_end]

        donor_start = -intronr_len - self.donor_exon_len
        donor_end = -intronr_len + self.donor_intron_len
        donor = x[donor_start: donor_end]

        donor_intron = x[-intronr_len + self.donor_intron_cut:]

        if donor[self.donor_exon_len:self.donor_exon_len + 2] != "GT":
            warnings.warn("None GT donor", UserWarning)
        if acceptor[self.acceptor_intron_len - 2:self.acceptor_intron_len] != "AG":
            warnings.warn("None AG donor", UserWarning)

        splits = {
            "acceptor_intron": acceptor_intron,
            "acceptor": acceptor,
            "exon": exon,
            "donor": donor,
            "donor_intron": donor_intron
        }

        if self.encode:
            return {k: encodeDNA([v]) for k, v in splits.items()}

        return splits



import numpy as np
from flask import Flask, jsonify, request
from keras import backend as K

from .mmsplice import MMSplice
from .utils import predict_deltaLogitPsi, predict_pathogenicity

app = Flask(__name__)

psi_model = None


@app.route('/create-model', methods=['POST'])
def create_model():
    global psi_model

    K.clear_session()
    psi_model = MMSplice(**{
        k: v
        for k, v in request.json.items()
        if v
    })

    # warms up the model
    psi_model.predict({
        'seq': "A" * 100,
        'intronl_len': 4,
        'intronr_len': 4
    })

    return jsonify(success=True)


@app.route('/psi-score', methods=['POST'])
def psi_score():
    global psi_model

    ref_scores, alt_scores = [
        np.matrix(
            psi_model.predict({
                "intronl_len": request.json['intronl_len'],
                "intronr_len": request.json['intronr_len'],
                "seq": request.json[seq]
            }).values
        )
        for seq in ['ref_seq', 'alt_seq']
    ]

    scores = np.hstack([ref_scores, alt_scores]).tolist()[0]

    scores.extend([
        predict_deltaLogitPsi(ref_scores, alt_scores)[0],
        int(predict_pathogenicity(ref_scores, alt_scores)[0])
    ])

    return ','.join(map(str, scores))



import numpy as np

bases = ['A', 'C', 'G', 'T']


def clip(x):
    return np.clip(x, 0.00001, 0.99999)


def logit(x):
    x = clip(x)
    return np.log(x) - np.log(1 - x)


def expit(x):
    return 1. / (1. + np.exp(-x))


def get_var_side(var):
    ''' Get exon variant side
    '''
    varstart, ref, alt, start, end, strand = var
    varend = varstart + len(ref) - 1
    # for insertion deletion, find the actual start of variant
    # e.g. A->AGG: start from G position, CA->CAGG:start from G position,
    # ATT->A: start from T position, CAT->CA: start from T position
    # For SNP var.POS is the actual mutation position
    if len(ref) != len(alt):
        # indels
        varstart = varstart + min(len(ref), len(alt))
    if strand == "+":
        if varstart < start:
            return "left"
        elif varend > end:
            return "right"
        else:
            return None
    else:
        if varstart < start:
            return "right"
        elif varend > end:
            return "left"
        else:
            return None


def onehot(seq):
    X = np.zeros((len(seq), len(bases)))
    for i, char in enumerate(seq):
        if char == "N":
            pass
        else:
            X[i, bases.index(char.upper())] = 1
    return X


def reverse_complement(dna):
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N'}
    return ''.join([complement[base] for base in dna[::-1]])


class Variant(object):

    def __init__(self,
                 CHROM,
                 POS,
                 REF,
                 ALT,
                 ID=None,
                 strand="*",
                 side=None):
        """ side: variant side of reference, e.g. splice juction.
        Sequence will be trimed or padded based on this for indels
        strand: specify if the variant is intepreted at any strand. e.g. in the case of splicing
        """
        self.CHROM = CHROM
        self.POS = POS
        self.ID = ID
        self.strand = strand
        self.REF = reverse_complement(REF.upper()) if strand == '-' else REF
        self.ALT = reverse_complement(ALT.upper()) if strand == '-' else ALT
        self.side = side
        self.len_diff = len(self.REF) - len(self.ALT)

    @property
    def is_snp(self):
        return len(self.REF) == 1 and len(self.ALT) == 1

    @property
    def to_dict(self):
        REF = reverse_complement(self.REF) if self.strand == '-' else self.REF
        ALT = reverse_complement(self.ALT) if self.strand == '-' else self.ALT
        return {
            'CHROM': self.CHROM,
            'POS': self.POS,
            'ID': self.ID,
            'REF': REF,
            'ALT': ALT,
            'STR': "%s:%s:%s:['%s']" % (self.CHROM, str(self.POS), REF, ALT)
        }

    def __repr__(self):
        return "Variant(CHROM={0}, POS={1}, REF={2}, ALT={3}, ID={4})".format(
            self.CHROM, self.POS, self.REF, self.ALT, self.ID)


class SpliceSite(object):
    ''' A splice site with flanking intron and exon sequence
    Args:
    order: order of splice site (donor or acceptor) in a transcript counted from 5' to 3'.
    variant: Variant class instance. Variant associated with this site.
    '''

    def __init__(self,
                 chrom,
                 start,
                 stop,
                 strand,
                 transcript_id,
                 gene_id,
                 biotype,
                 order=None,
                 variant=None,
                 encode=True):
        self.chrom = chrom
        self.grange = (start, stop)
        self.strand = strand
        self.transcriptID = transcript_id
        self.geneID = gene_id
        self.biotype = biotype
        self.order = order
        self.seq = None
        self.variant = variant
        self.encode = encode

    def get_seq(self, fasta):
        seq = fasta.get_seq(self.chrom, self.grange, self.strand).upper()
        return onehot(seq) if self.encode else seq



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'setuptools<=39.1.0',
    'sklearn',
    'tensorflow',
    'kipoi',
    'pandas',
    'concise',
    'cyvcf2==0.9.0',
    'gffutils',
    'pyfaidx',
    'tqdm',
    'click',
    'flask',
    'gevent'
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Jun Cheng",
    author_email='chengju@in.tum.de',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        # "Programming Language :: Python :: 2",
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="Predict splicing variant effect from VCF",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    # package_data={
    #     'mmsplice.models':
    #     ['Acceptor.h5',
    #     'Donor.h5',
    #     'Exon.h5',
    #     'Intron3.h5',
    #     'Intron5.h5']
    # }, # Done with MANIFEST.in

    entry_points='''
        [console_scripts]
        mmsplice=mmsplice.main:cli
    ''',

    include_package_data=True,
    keywords='mmsplice',
    name='mmsplice',
    packages=find_packages(include=['mmsplice']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/s6juncheng/mmsplice',
    version='0.2.5',
    zip_safe=False,
)



"""
Accept GenomicInterval object, with attributes chrom, start, end, strand (optional).
In case of unstranded data, interval.strand == '.'
"""
import math
import random
import warnings


class Interval(object):
    ''' Compatible with pybedtools Interval
    '''

    def __init__(self,
                 chrom=None,
                 start=None,
                 end=None,
                 strand="*",
                 name=None):
        self.chrom = chrom
        assert start <= end, "start position smaller or equal to the end position"
        self.start = start
        self.end = end
        self.name = name
        self.strand = strand

    def __repr__(self):
        return "{0}, {1}:{2}-{3},{4}".format(
            self.name, self.chrom, self.start, self.end, self.strand)


class IntervalTree(object):
    def __init__(self):
        self.chroms = {}

    def insert(self, interval):
        # This interval is the interval to construct the tree, e.g. gtf annotations
        if interval.chrom in self.chroms:
            self.chroms[interval.chrom] = self.chroms[interval.chrom].insert(
                interval)
        else:
            self.chroms[interval.chrom] = IntervalNode(interval)

    def intersect(self, interval, ignore_strand=False):
        # This interval from the query
        if interval.chrom in self.chroms:
            yield from self.chroms[interval.chrom].intersect(
                interval, ignore_strand=ignore_strand)
        else:
            warnings.warn("Interval chromosome name no match", UserWarning)

    def traverse(self):
        for item in self.chroms.itervalues():
            yield from item.traverse()


class IntervalNode(object):
    def __init__(self, interval):
        self.priority = math.ceil(
            (-1. / math.log(.5)) * math.log(-1. / (random.uniform(0, 1) - 1)))
        self.interval = interval
        self.start = interval.start
        self.end = interval.end
        self.strand = interval.strand
        self.maxend = self.end
        self.minend = self.end
        self.left = None
        self.right = None

    def insert(self, interval):
        root = self
        if interval.start > self.start:
            # insert to right tree
            if self.right:
                self.right = self.right.insert(interval)
            else:
                self.right = IntervalNode(interval)
            # rebalance tree
            if self.priority < self.right.priority:
                root = self.rotateleft()
        else:
            # insert to left tree
            if self.left:
                self.left = self.left.insert(interval)
            else:
                self.left = IntervalNode(interval)
            # rebalance tree
            if self.priority < self.left.priority:
                root = self.rotateright()
        if root.right and root.left:
            root.maxend = max(root.end, root.right.maxend, root.left.maxend)
            root.minend = min(root.end, root.right.minend, root.left.minend)
        elif root.right:
            root.maxend = max(root.end, root.right.maxend)
            root.minend = min(root.end, root.right.minend)
        elif root.left:
            root.maxend = max(root.end, root.left.maxend)
            root.minend = min(root.end, root.left.minend)
        return root

    def rotateright(self):
        root = self.left
        self.left = self.left.right
        root.right = self
        if self.right and self.left:
            self.maxend = max(self.end, self.right.maxend, self.left.maxend)
            self.minend = min(self.end, self.right.minend, self.left.minend)
        elif self.right:
            self.maxend = max(self.end, self.right.maxend)
            self.minend = min(self.end, self.right.minend)
        elif self.left:
            self.maxend = max(self.end, self.left.maxend)
            self.minend = min(self.end, self.left.minend)
        return root

    def rotateleft(self):
        root = self.right
        self.right = self.right.left
        root.left = self
        if self.right and self.left:
            self.maxend = max(self.end, self.right.maxend, self.left.maxend)
            self.minend = min(self.end, self.right.minend, self.left.minend)
        elif self.right:
            self.maxend = max(self.end, self.right.maxend)
            self.minend = min(self.end, self.right.minend)
        elif self.left:
            self.maxend = max(self.end, self.left.maxend)
            self.minend = min(self.end, self.left.minend)
        return root

    def intersect(self, interval, ignore_strand=False):
        # unstranded data, not going to compare strand
        if interval.strand == '*' \
           or ignore_strand \
           or interval.strand == self.strand:
            if interval.start <= self.end and interval.end >= self.start:
                yield self

        if self.left and interval.start <= self.left.maxend:
            yield from self.left.intersect(
                interval, ignore_strand=ignore_strand)
        if self.right and interval.end >= self.start:
            yield from self.right.intersect(
                interval, ignore_strand=ignore_strand)

    def traverse(self):
        if self.left:
            yield from self.left.traverse()
        yield self
        if self.right:
            yield from self.right.traverse()




import pandas as pd
import numpy as np
from sklearn.externals import joblib
from pkg_resources import resource_filename
LINEAR_MODEL = joblib.load(resource_filename(
    'mmsplice', 'models/linear_model.pkl'))
LOGISTIC_MODEL = joblib.load(resource_filename(
    'mmsplice', 'models/Pathogenicity.pkl'))


def max_varEff(df):
    """ Summarize largest absolute effect per variant across all affected exons.
    Args:
        df: result of `predict_all_table`
    """
    if isinstance(df, str):
        df = pd.read_csv(df, index_col=0)

    ref_list = ['mmsplice_ref_acceptorIntron',
                'mmsplice_ref_acceptor',
                'mmsplice_ref_exon',
                'mmsplice_ref_donor',
                'mmsplice_ref_donorIntron']
    alt_list = ['mmsplice_alt_acceptorIntron',
                'mmsplice_alt_acceptor',
                'mmsplice_alt_exon',
                'mmsplice_alt_donor',
                'mmsplice_alt_donorIntron']

    if 'mmsplice_dlogitPsi' not in df.columns:
        X = df[alt_list].values - df[ref_list].values
        X = transform(X)
        df['mmsplice_dlogitPsi'] = LINEAR_MODEL.predict(X)

    dfMax = df.groupby(['ID'], as_index=False).agg(
        {'mmsplice_dlogitPsi': lambda x: max(x, key=abs)})

    dfMax = dfMax.merge(df, how='left', on=['ID', 'mmsplice_dlogitPsi'])
    dfMax = dfMax.drop_duplicates(subset=['ID', 'mmsplice_dlogitPsi'])
    # dfMax = dfMax.drop("mmsplice_dlogitPsi", axis=1)
    return dfMax


def _not_close0(arr):
    return ~np.isclose(arr, 0)


def _and_not_close0(x, y):
    return np.logical_and(_not_close0(x), _not_close0(y))


def transform(X, region_only=False):
    ''' Make interaction terms for the overlapping prediction region
    Args:
        X: modular prediction. Shape (, 5)
        region_only: only interaction terms with indicator function on overlapping
    '''
    exon_overlap = np.logical_or(
        _and_not_close0(X[:, 1], X[:, 2]),
        _and_not_close0(X[:, 2], X[:, 3])
    )
    acceptor_intron_overlap = _and_not_close0(X[:, 0], X[:, 1])
    donor_intron_overlap = _and_not_close0(X[:, 3], X[:, 4])

    if not region_only:
        exon_overlap = X[:, 2] * exon_overlap
        donor_intron_overlap = X[:, 4] * donor_intron_overlap
        acceptor_intron_overlap = X[:, 0] * acceptor_intron_overlap

    return np.hstack([
        X,
        exon_overlap.reshape(-1, 1),
        donor_intron_overlap.reshape(-1, 1),
        acceptor_intron_overlap.reshape(-1, 1)
    ])


def predict_deltaLogitPsi(X_ref, X_alt):
    return LINEAR_MODEL.predict(transform(X_alt - X_ref, region_only=False))


def predict_pathogenicity(X_ref, X_alt):
    X = transform(X_alt - X_ref, region_only=True)
    X = np.concatenate([X_ref, X_alt, X[:, -3:]], axis=-1)
    return LOGISTIC_MODEL.predict(X)
